﻿//******************************************************************************************************************************************************************************************//
// Copyright (c) 2021 abergs (https://github.com/abergs/fido2-net-lib)                                                                                                                      //                        
//                                                                                                                                                                                          //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),                                       //
// to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,   //
// and to permit persons to whom the Software is furnished to do so, subject to the following conditions:                                                                                   //
//                                                                                                                                                                                          //
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.                                                           //
//                                                                                                                                                                                          //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,                                      //
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,                            //
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               //
//                                                                                                                                                                                          //
//******************************************************************************************************************************************************************************************//
// Copyright (c) 2024 redhook (adfsmfa@gmail.com)                                                                                                                                    //                        
//                                                                                                                                                                                          //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),                                       //
// to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,   //
// and to permit persons to whom the Software is furnished to do so, subject to the following conditions:                                                                                   //
//                                                                                                                                                                                          //
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.                                                           //
//                                                                                                                                                                                          //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,                                      //
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,                            //
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               //
//                                                                                                                                                                                          //
//                                                                                                                                                             //
// https://github.com/neos-sdi/adfsmfa                                                                                                                                                      //
//                                                                                                                                                                                          //
//******************************************************************************************************************************************************************************************//
using Neos.IdentityServer.MultiFactor.WebAuthN.Library.ASN;
using Neos.IdentityServer.MultiFactor.WebAuthN.Library.Cbor;
using Neos.IdentityServer.MultiFactor.WebAuthN.Objects;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;

namespace Neos.IdentityServer.MultiFactor.WebAuthN.AttestationFormat
{
    internal sealed class Tpm : AttestationVerifier
    {
        public static readonly List<string> TPMManufacturers = new List<string>
        {
            "id:FFFFF1D0", // FIDO testing TPM
            // From https://trustedcomputinggroup.org/wp-content/uploads/TCG-TPM-Vendor-ID-Registry-Version-1.02-Revision-1.00.pdf
            "id:414D4400", // 'AMD' AMD
            "id:41544D4C", // 'ATML' Atmel
            "id:4252434D", // 'BRCM' Broadcom
            "id:4353434F", // 'CSCO' Cisco
            "id:464C5953", // 'FLYS' Flyslice Technologies
            "id:48504500", // 'HPE' HPE
            "id:49424d00", // 'IBM' IBM
            "id:49465800", // 'IFX' Infinion
            "id:494E5443", // 'INTC' Intel
            "id:4C454E00", // 'LEN' Lenovo
            "id:4D534654", // 'MSFT' Microsoft
            "id:4E534D20", // 'NSM' National Semiconductor
            "id:4E545A00", // 'NTZ' Nationz 
            "id:4E544300", // 'NTC' Nuvoton Technology
            "id:51434F4D", // 'QCOM' Qualcomm
            "id:534D5343", // 'SMSC' SMSC
            "id:53544D20", // 'STM ' ST Microelectronics
            "id:534D534E", // 'SMSN' Samsung
            "id:534E5300", // 'SNS' Sinosun
            "id:54584E00", // 'TXN' Texas Instruments
            "id:57454300", // 'WEC' Winbond
            "id:524F4343", // 'ROCC' Fuzhou Rockchip
            "id:474F4F47", // 'GOOG' Google
        };

        public override (AttestationType, X509Certificate2[]) Verify()
        {
            // 1. Verify that attStmt is valid CBOR conforming to the syntax defined above and perform CBOR decoding on it to extract the contained fields.
            // (handled in base class)
            if (null == Sig || CBORType.ByteString != Sig.Type || 0 == Sig.GetByteString().Length)
                throw new VerificationException("Invalid TPM attestation signature");

            if ("2.0" != attStmt["ver"].AsString())
                throw new VerificationException("FIDO2 only supports TPM 2.0");

            // 2. Verify that the public key specified by the parameters and unique fields of pubArea
            // is identical to the credentialPublicKey in the attestedCredentialData in authenticatorData
            PubArea pubArea = null;
            if (null != attStmt["pubArea"] &&
                CBORType.ByteString == attStmt["pubArea"].Type &&
                0 != attStmt["pubArea"].GetByteString().Length)
            {
                pubArea = new PubArea(attStmt["pubArea"].GetByteString());
            }

            if (null == pubArea || null == pubArea.Unique || 0 == pubArea.Unique.Length)
                throw new VerificationException("Missing or malformed pubArea");

            var coseKty = CredentialPublicKey[CBORObject.FromObject(COSE.KeyCommonParameter.KeyType)].AsInt32();
            if (3 == coseKty) // RSA
            {
                var coseMod = CredentialPublicKey[CBORObject.FromObject(COSE.KeyTypeParameter.N)].GetByteString(); // modulus 
                var coseExp = CredentialPublicKey[CBORObject.FromObject(COSE.KeyTypeParameter.E)].GetByteString(); // exponent

                if (!coseMod.ToArray().SequenceEqual(pubArea.Unique.ToArray()))
                    throw new VerificationException("Public key mismatch between pubArea and credentialPublicKey");
                if ((coseExp[0] + (coseExp[1] << 8) + (coseExp[2] << 16)) != pubArea.Exponent)
                    throw new VerificationException("Public key exponent mismatch between pubArea and credentialPublicKey");
            }
            else if (2 == coseKty) // ECC
            {
                var curve = CredentialPublicKey[CBORObject.FromObject(COSE.KeyTypeParameter.Crv)].AsInt32();
                var X = CredentialPublicKey[CBORObject.FromObject(COSE.KeyTypeParameter.X)].GetByteString();
                var Y = CredentialPublicKey[CBORObject.FromObject(COSE.KeyTypeParameter.Y)].GetByteString();

                if (pubArea.EccCurve != CoseCurveToTpm[curve])
                    throw new VerificationException("Curve mismatch between pubArea and credentialPublicKey");
                if (!pubArea.ECPoint.X.SequenceEqual(X))
                    throw new VerificationException("X-coordinate mismatch between pubArea and credentialPublicKey");
                if (!pubArea.ECPoint.Y.SequenceEqual(Y))
                    throw new VerificationException("Y-coordinate mismatch between pubArea and credentialPublicKey");
            }
            // 3. Concatenate authenticatorData and clientDataHash to form attToBeSigned
            // See Data field of base class

            // 4. Validate that certInfo is valid
            CertInfo certInfo = null;
            if (null != attStmt["certInfo"] &&
                CBORType.ByteString == attStmt["certInfo"].Type &&
                0 != attStmt["certInfo"].GetByteString().Length)
            {
                certInfo = new CertInfo(attStmt["certInfo"].GetByteString());
            }

            if (null == certInfo)
                throw new VerificationException("CertInfo invalid parsing TPM format attStmt");

            // 4a. Verify that magic is set to TPM_GENERATED_VALUE
            // Handled in CertInfo constructor, see CertInfo.Magic

            // 4b. Verify that type is set to TPM_ST_ATTEST_CERTIFY
            // Handled in CertInfo constructor, see CertInfo.Type

            // 4c. Verify that extraData is set to the hash of attToBeSigned using the hash algorithm employed in "alg"
            if (null == Alg || true != Alg.IsNumber)
                throw new VerificationException("Invalid TPM attestation algorithm");

            using (var hasher = CryptoUtils.GetHasher(CryptoUtils.HashAlgFromCOSEAlg(Alg.AsInt32())))
            {
                if (!hasher.ComputeHash(Data).SequenceEqual(certInfo.ExtraData))
                    throw new VerificationException("Hash value mismatch extraData and attToBeSigned");
            }

            // 4d. Verify that attested contains a TPMS_CERTIFY_INFO structure, whose name field contains a valid Name for pubArea, as computed using the algorithm in the nameAlg field of pubArea 
            using (var hasher = CryptoUtils.GetHasher(CryptoUtils.HashAlgFromCOSEAlg(certInfo.Alg)))
            {
                if (false == hasher.ComputeHash(pubArea.Raw).SequenceEqual(certInfo.AttestedName))
                    throw new VerificationException("Hash value mismatch attested and pubArea");
            }

            // 4e. Note that the remaining fields in the "Standard Attestation Structure" [TPMv2-Part1] section 31.2, i.e., qualifiedSigner, clockInfo and firmwareVersion are ignored. These fields MAY be used as an input to risk engines.

            // 5. If x5c is present, this indicates that the attestation type is not ECDAA
            if (null != X5c && CBORType.Array == X5c.Type && 0 != X5c.Count)
            {
                if (null == X5c.Values || 0 == X5c.Values.Count ||
                    CBORType.ByteString != X5c.Values.First().Type ||
                    0 == X5c.Values.First().GetByteString().Length)
                {
                    throw new VerificationException("Malformed x5c in TPM attestation");
                }

                // 5a. Verify the sig is a valid signature over certInfo using the attestation public key in aikCert with the algorithm specified in alg.
                var aikCert = new X509Certificate2(X5c.Values.First().GetByteString());

                var cpk = new CredentialPublicKey(aikCert, Alg.AsInt32());
                if (true != cpk.Verify(certInfo.Raw, Sig.GetByteString()))
                    throw new VerificationException("Bad signature in TPM with aikCert");

                // 5b. Verify that aikCert meets the TPM attestation statement certificate requirements
                // https://www.w3.org/TR/webauthn/#tpm-cert-requirements
                // 5bi. Version MUST be set to 3
                if (3 != aikCert.Version)
                    throw new VerificationException("aikCert must be V3");

                // 5bii. Subject field MUST be set to empty - they actually mean subject name
                if (0 != aikCert.SubjectName.Name.Length)
                    throw new VerificationException("aikCert subject must be empty");

                // 5biii. The Subject Alternative Name extension MUST be set as defined in [TPMv2-EK-Profile] section 3.2.9.
                // https://www.w3.org/TR/webauthn/#tpm-cert-requirements
                (string tpmManufacturer, string tpmModel, string tpmVersion) = SANFromAttnCertExts(aikCert.Extensions);

                // From https://www.trustedcomputinggroup.org/wp-content/uploads/Credential_Profile_EK_V2.0_R14_published.pdf
                // "The issuer MUST include TPM manufacturer, TPM part number and TPM firmware version, using the directoryName 
                // form within the GeneralName structure. The ASN.1 encoding is specified in section 3.1.2 TPM Device 
                // Attributes. In accordance with RFC 5280[11], this extension MUST be critical if subject is empty 
                // and SHOULD be non-critical if subject is non-empty"

                // Best I can figure to do for now?
                if (string.Empty == tpmManufacturer ||
                    string.Empty == tpmModel ||
                    string.Empty == tpmVersion)
                {
                    throw new VerificationException("SAN missing TPMManufacturer, TPMModel, or TPMVersion from TPM attestation certificate");
                }

                if (false == TPMManufacturers.Contains(tpmManufacturer))
                    throw new VerificationException("Invalid TPM manufacturer found parsing TPM attestation");

                // 5biiii. The Extended Key Usage extension MUST contain the "joint-iso-itu-t(2) internationalorganizations(23) 133 tcg-kp(8) tcg-kp-AIKCertificate(3)" OID.
                // OID is 2.23.133.8.3
                var EKU = EKUFromAttnCertExts(aikCert.Extensions, "2.23.133.8.3");
                if (!EKU)
                    throw new VerificationException("aikCert EKU missing tcg-kp-AIKCertificate OID");

                // 5biiiii. The Basic Constraints extension MUST have the CA component set to false.
                if (IsAttnCertCACert(aikCert.Extensions))
                    throw new VerificationException("aikCert Basic Constraints extension CA component must be false");

                // 5biiiiii. An Authority Information Access (AIA) extension with entry id-ad-ocsp and a CRL Distribution Point extension [RFC5280] 
                // are both OPTIONAL as the status of many attestation certificates is available through metadata services. See, for example, the FIDO Metadata Service [FIDOMetadataService].
                var trustPath = X5c.Values
                    .Select(x => new X509Certificate2(x.GetByteString()))
                    .ToArray();

                // 5c. If aikCert contains an extension with OID 1.3.6.1.4.1.45724.1.1.4 (id-fido-gen-ce-aaguid) verify that the value of this extension matches the aaguid in authenticatorData
                var aaguid = AaguidFromAttnCertExts(aikCert.Extensions);
                if ((null != aaguid) &&
                    (!aaguid.SequenceEqual(Guid.Empty.ToByteArray())) &&
                    (0 != AttestedCredentialData.FromBigEndian(aaguid).CompareTo(AuthData.AttestedCredentialData.AaGuid)))
                    throw new VerificationException(string.Format("aaguid malformed, expected {0}, got {1}", AuthData.AttestedCredentialData.AaGuid, new Guid(aaguid)));

                return (AttestationType.AttCa, trustPath);
            }
            // If ecdaaKeyId is present, then the attestation type is ECDAA
            else if (null != EcdaaKeyId)
            {
                // Perform ECDAA-Verify on sig to verify that it is a valid signature over certInfo
                // https://www.w3.org/TR/webauthn/#biblio-fidoecdaaalgorithm
                throw new VerificationException("ECDAA support for TPM attestation is not yet implemented");
                // If successful, return attestation type ECDAA and the identifier of the ECDAA-Issuer public key ecdaaKeyId.
                //attnType = AttestationType.ECDAA;
                //trustPath = ecdaaKeyId;
            }
            else
            {
                throw new VerificationException("Neither x5c nor ECDAA were found in the TPM attestation statement");
            }
        }
        private static readonly Dictionary<int, TpmEccCurve> CoseCurveToTpm = new Dictionary<int, TpmEccCurve>
        {
            { 1, TpmEccCurve.TPM_ECC_NIST_P256},
            { 2, TpmEccCurve.TPM_ECC_NIST_P384},
            { 3, TpmEccCurve.TPM_ECC_NIST_P521}
        };
        private static (string, string, string) SANFromAttnCertExts(X509ExtensionCollection extensions)
        {
            string tpmManufacturer = string.Empty,
                tpmModel = string.Empty,
                tpmVersion = string.Empty;

            var foundSAN = false;

            foreach (var extension in extensions)
            {
                if (extension.Oid.Value.Equals("2.5.29.17")) // subject alternative name
                {
                    if (0 == extension.RawData.Length)
                        throw new VerificationException("SAN missing from TPM attestation certificate");

                    foundSAN = true;

                    var subjectAlternativeName = AsnElt.Decode(extension.RawData);
                    subjectAlternativeName.CheckConstructed();
                    subjectAlternativeName.CheckTag(AsnElt.SEQUENCE);
                    subjectAlternativeName.CheckNumSubMin(1);

                    var generalName = subjectAlternativeName.Sub.FirstOrDefault(o => o.TagClass == AsnElt.CONTEXT && o.TagValue == AsnElt.OCTET_STRING);

                    if (generalName != null)
                    {
                        generalName.CheckConstructed();
                        generalName.CheckNumSub(1);

                        var nameSequence = generalName.GetSub(0);
                        nameSequence.CheckConstructed();
                        nameSequence.CheckTag(AsnElt.SEQUENCE);
                        nameSequence.CheckNumSubMin(1);

                        /*
                         
                        Per Trusted Computing Group Endorsement Key Credential Profile section 3.2.9:

                        "The issuer MUST include TPM manufacturer, TPM part number and TPM firmware version, using the directoryName-form within the GeneralName structure. The ASN.1 encoding is specified in section 3.1.2 TPM Device Attributes."

                        An example is provided in document section A.1 Example 1:

                                    // SEQUENCE
                                    30 49
                                         // SET
                                         31 16
                                             // SEQUENCE
                                             30 14
                                                 // OBJECT IDENTIFER tcg-at-tpmManufacturer (2.23.133.2.1)
                                                 06 05 67 81 05 02 01
                                                 // UTF8 STRING id:54434700 (TCG)
                                                 0C 0B 69 64 3A 35 34 34 33 34 37 30 30
                                        // SET
                                        31 17
                                            // SEQUENCE
                                            30 15
                                                // OBJECT IDENTIFER tcg-at-tpmModel (2.23.133.2.2)
                                                06 05 67 81 05 02 02
                                                // UTF8 STRING ABCDEF123456
                                                0C 0C 41 42 43 44 45 46 31 32 33 34 35 36
                                        // SET
                                        31 16
                                            // SEQUENCE
                                            30 14
                                                // OBJECT IDENTIFER tcg-at-tpmVersion (2.23.133.2.3)
                                                06 05 67 81 05 02 03
                                                // UTF8 STRING id:00010023
                                                0C 0B 69 64 3A 30 30 30 31 30 30 32 33

                        Some TPM implementations place each device attributes SEQUENCE within a single SET instead of each in its own SET.

                        This detects this condition and repacks each devices attributes SEQUENCE into its own SET to conform with TCG spec.

                         */

                        var deviceAttributes = nameSequence.Sub;
                        if (1 != deviceAttributes.FirstOrDefault().Sub.Length)
                        {
                            deviceAttributes = deviceAttributes.FirstOrDefault().Sub.Select(o => AsnElt.Make(AsnElt.SET, o)).ToArray();
                        }

                        foreach (AsnElt propertySet in deviceAttributes)
                        {
                            propertySet.CheckTag(AsnElt.SET);
                            propertySet.CheckNumSub(1);

                            var propertySequence = propertySet.GetSub(0);
                            propertySequence.CheckTag(AsnElt.SEQUENCE);
                            propertySequence.CheckNumSub(2);

                            var propertyOid = propertySequence.GetSub(0);
                            propertyOid.CheckTag(AsnElt.OBJECT_IDENTIFIER);
                            propertyOid.CheckPrimitive();

                            var propertyValue = propertySequence.GetSub(1);
                            propertyValue.CheckTag(AsnElt.UTF8String);
                            propertyValue.CheckPrimitive();

                            switch (propertyOid.GetOID())
                            {
                                case ("2.23.133.2.1"):
                                    tpmManufacturer = propertyValue.GetString();
                                    break;
                                case ("2.23.133.2.2"):
                                    tpmModel = propertyValue.GetString();
                                    break;
                                case ("2.23.133.2.3"):
                                    tpmVersion = propertyValue.GetString();
                                    break;
                                default:
                                    continue;
                            }
                        }
                    }

                    break;
                }
            }

            if (!foundSAN)
                throw new VerificationException("SAN missing from TPM attestation certificate");

            return (tpmManufacturer, tpmModel, tpmVersion);
        }

        private static bool EKUFromAttnCertExts(X509ExtensionCollection exts, string expectedEnhancedKeyUsages)
        {
            foreach (var ext in exts)
            {
                if (ext.Oid.Value.Equals("2.5.29.37") && ext is X509EnhancedKeyUsageExtension enhancedKeyUsageExtension)
                {
                    foreach (var oid in enhancedKeyUsageExtension.EnhancedKeyUsages)
                    {
                        if (expectedEnhancedKeyUsages.Equals(oid.Value))
                            return true;
                    }

                }
            }
            return false;
        }
    }

    public enum TpmEccCurve : ushort
    {
        // TCG TPM Rev 2.0, part 2, structures, section 6.4, TPM_ECC_CURVE
        TPM_ECC_NONE,       // 0x0000
        TPM_ECC_NIST_P192,  // 0x0001
        TPM_ECC_NIST_P224,  // 0x0002
        TPM_ECC_NIST_P256,  // 0x0003
        TPM_ECC_NIST_P384,  // 0x0004
        TPM_ECC_NIST_P521,  // 0x0005  
        TPM_ECC_BN_P256,    // 0x0010 curve to support ECDAA
        TPM_ECC_BN_P638,    // 0x0011 curve to support ECDAA
        TPM_ECC_SM2_P256    // 0x0020 
    }
    public enum TpmAlg : ushort
    {
        // TCG TPM Rev 2.0, part 2, structures, section 6.3, TPM_ALG_ID
        TPM_ALG_ERROR, // 0
        TPM_ALG_RSA, // 1
        TPM_ALG_SHA1 = 4, // 4
        TPM_ALG_HMAC, // 5
        TPM_ALG_AES, // 6
        TPM_ALG_MGF1, // 7
        TPM_ALG_KEYEDHASH, // 8
        TPM_ALG_XOR = 0xA, // A
        TPM_ALG_SHA256, // B
        TPM_ALG_SHA384, // C
        TPM_ALG_SHA512, // D
        TPM_ALG_NULL = 0x10, // 10
        TPM_ALG_SM3_256 = 0x12, // 12
        TPM_ALG_SM4, // 13
        TPM_ALG_RSASSA, // 14
        TPM_ALG_RSAES, // 15
        TPM_ALG_RSAPSS, // 16
        TPM_ALG_OAEP, // 17
        TPM_ALG_ECDSA, // 18
        TPM_ALG_ECDH, // 19
        TPM_ALG_ECDAA, // 1A
        TPM_ALG_SM2, // 1B
        TPM_ALG_ECSCHNORR, // 1C
        TPM_ALG_ECMQV, // 1D
        TPM_ALG_KDF1_SP800_56A = 0x20,
        TPM_ALG_KDF2, // 21
        TPM_ALG_KDF1_SP800_108, // 22
        TPM_ALG_ECC, // 23
        TPM_ALG_SYMCIPHER = 0x25,
        TPM_ALG_CAMELLIA, // 26
        TPM_ALG_CTR = 0x40,
        TPM_ALG_OFB, // 41
        TPM_ALG_CBC, // 42 
        TPM_ALG_CFB, // 43
        TPM_ALG_ECB // 44
    };
    // TPMS_ATTEST, TPMv2-Part2, section 10.12.8
    public class CertInfo
    {
        private static readonly Dictionary<TpmAlg, ushort> tpmAlgToDigestSizeMap = new Dictionary<TpmAlg, ushort>
        {
            {TpmAlg.TPM_ALG_SHA1,   (160/8) },
            {TpmAlg.TPM_ALG_SHA256, (256/8) },
            {TpmAlg.TPM_ALG_SHA384, (384/8) },
            {TpmAlg.TPM_ALG_SHA512, (512/8) }
        };

        public static (ushort size, byte[] name) NameFromTPM2BName(byte[] ab, ref int offset)
        {
            // TCG TPM Rev 2.0, part 2, structures, section 10.5.3, TPM2B_NAME
            // This buffer holds a Name for any entity type. 
            // The type of Name in the structure is determined by context and the size parameter. 
            var totalBytes = AuthDataHelper.GetSizedByteArray(ab, ref offset, 2);
            ushort totalSize = 0;
            if (null != totalBytes)
            {
                totalSize = BitConverter.ToUInt16(totalBytes.ToArray().Reverse().ToArray(), 0);
            }
            ushort size = 0;
            var bytes = AuthDataHelper.GetSizedByteArray(ab, ref offset, 2);
            if (null != bytes)
            {
                size = BitConverter.ToUInt16(bytes.ToArray().Reverse().ToArray(), 0);
            }
            // If size is four, then the Name is a handle. 
            if (4 == size)
                throw new VerificationException("Unexpected handle in TPM2B_NAME");
            // If size is zero, then no Name is present. 
            if (0 == size)
                throw new VerificationException("Unexpected no name found in TPM2B_NAME");
            // Otherwise, the size shall be the size of a TPM_ALG_ID plus the size of the digest produced by the indicated hash algorithm.
            byte[] name = null;
            if (Enum.IsDefined(typeof(TpmAlg), size))
            {
                var tpmalg = (TpmAlg)size;
                if (tpmAlgToDigestSizeMap.ContainsKey(tpmalg))
                {
                    name = AuthDataHelper.GetSizedByteArray(ab, ref offset, tpmAlgToDigestSizeMap[tpmalg]);
                }
                else
                {
                    throw new VerificationException("TPM_ALG_ID found in TPM2B_NAME not acceptable hash algorithm");
                }
            }
            else
            {
                throw new VerificationException("Invalid TPM_ALG_ID found in TPM2B_NAME");
            }

            if (totalSize != bytes.Length + name.Length)
                throw new VerificationException("Unexpected extra bytes found in TPM2B_NAME");
            return (size, name);
        }

        public CertInfo(byte[] certInfo)
        {
            if (null == certInfo || 0 == certInfo.Length)
                throw new VerificationException("Malformed certInfo bytes");
            Raw = certInfo;
            var offset = 0;
            Magic = AuthDataHelper.GetSizedByteArray(certInfo, ref offset, 4);
            if (0xff544347 != BitConverter.ToUInt32(Magic.ToArray().Reverse().ToArray(), 0))
                throw new VerificationException("Bad magic number " + BitConverter.ToString(Magic).Replace("-", ""));
            Type = AuthDataHelper.GetSizedByteArray(certInfo, ref offset, 2);
            if (0x8017 != BitConverter.ToUInt16(Type.ToArray().Reverse().ToArray(), 0))
                throw new VerificationException("Bad structure tag " + BitConverter.ToString(Type).Replace("-", ""));
            QualifiedSigner = AuthDataHelper.GetSizedByteArray(certInfo, ref offset);
            ExtraData = AuthDataHelper.GetSizedByteArray(certInfo, ref offset);
            if (null == ExtraData || 0 == ExtraData.Length)
                throw new VerificationException("Bad extraData in certInfo");
            Clock = AuthDataHelper.GetSizedByteArray(certInfo, ref offset, 8);
            ResetCount = AuthDataHelper.GetSizedByteArray(certInfo, ref offset, 4);
            RestartCount = AuthDataHelper.GetSizedByteArray(certInfo, ref offset, 4);
            Safe = AuthDataHelper.GetSizedByteArray(certInfo, ref offset, 1);
            FirmwareVersion = AuthDataHelper.GetSizedByteArray(certInfo, ref offset, 8);
            var (size, name) = NameFromTPM2BName(certInfo, ref offset);
            Alg = size; // TPM_ALG_ID
            AttestedName = name;
            AttestedQualifiedNameBuffer = AuthDataHelper.GetSizedByteArray(certInfo, ref offset);
            if (certInfo.Length != offset)
                throw new VerificationException("Leftover bits decoding certInfo");
        }
        public byte[] Raw { get; private set; }
        public byte[] Magic { get; private set; }
        public byte[] Type { get; private set; }
        public byte[] QualifiedSigner { get; private set; }
        public byte[] ExtraData { get; private set; }
        public byte[] Clock { get; private set; }
        public byte[] ResetCount { get; private set; }
        public byte[] RestartCount { get; private set; }
        public byte[] Safe { get; private set; }
        public byte[] FirmwareVersion { get; private set; }
        public ushort Alg { get; private set; }
        public byte[] AttestedName { get; private set; }
        public byte[] AttestedQualifiedNameBuffer { get; private set; }
    }
    // TPMT_PUBLIC, TPMv2-Part2, section 12.2.4
    public class PubArea
    {
        public PubArea(byte[] pubArea)
        {
            Raw = pubArea;
            var offset = 0;

            // TPMI_ALG_PUBLIC
            Type = AuthDataHelper.GetSizedByteArray(pubArea, ref offset, 2);
            var tpmalg = (TpmAlg)Enum.Parse(typeof(TpmAlg), BitConverter.ToUInt16(Type.Reverse().ToArray(), 0).ToString());

            // TPMI_ALG_HASH 
            Alg = AuthDataHelper.GetSizedByteArray(pubArea, ref offset, 2);

            // TPMA_OBJECT, attributes that, along with type, determine the manipulations of this object 
            Attributes = AuthDataHelper.GetSizedByteArray(pubArea, ref offset, 4);

            // TPM2B_DIGEST, optional policy for using this key, computed using the alg of the object
            Policy = AuthDataHelper.GetSizedByteArray(pubArea, ref offset);

            // TPMU_PUBLIC_PARMS
            Symmetric = null;
            Scheme = null;

            if (TpmAlg.TPM_ALG_KEYEDHASH == tpmalg)
            {
                throw new VerificationException("TPM_ALG_KEYEDHASH not yet supported");
            }
            if (TpmAlg.TPM_ALG_SYMCIPHER == tpmalg)
            {
                throw new VerificationException("TPM_ALG_SYMCIPHER not yet supported");
            }

            // TPMS_ASYM_PARMS, for TPM_ALG_RSA and TPM_ALG_ECC
            if (TpmAlg.TPM_ALG_RSA == tpmalg || TpmAlg.TPM_ALG_ECC == tpmalg)
            {
                Symmetric = AuthDataHelper.GetSizedByteArray(pubArea, ref offset, 2);
                Scheme = AuthDataHelper.GetSizedByteArray(pubArea, ref offset, 2);
            }

            // TPMI_RSA_KEY_BITS, number of bits in the public modulus 
            KeyBits = null;
            // The public exponent, a prime number greater than 2. When zero, indicates that the exponent is the default of 2^16 + 1 
            Exponent = 0;

            if (TpmAlg.TPM_ALG_RSA == tpmalg)
            {
                KeyBits = AuthDataHelper.GetSizedByteArray(pubArea, ref offset, 2);
                var tmp = AuthDataHelper.GetSizedByteArray(pubArea, ref offset, 4);
                if (null != tmp)
                {
                    Exponent = BitConverter.ToUInt32(tmp.ToArray(), 0);
                    if (0 == Exponent) Exponent = Convert.ToUInt32(Math.Pow(2, 16) + 1);
                }
            }

            // TPMI_ECC_CURVE
            CurveID = null;

            // TPMT_KDF_SCHEME, an optional key derivation scheme for generating a symmetric key from a Z value 
            // If the kdf  parameter associated with curveID is not TPM_ALG_NULL then this is required to be NULL. 
            // NOTE There are currently no commands where this parameter has effect and, in the reference code, this field needs to be set to TPM_ALG_NULL. 
            KDF = null;

            if (TpmAlg.TPM_ALG_ECC == tpmalg)
            {
                CurveID = AuthDataHelper.GetSizedByteArray(pubArea, ref offset, 2);
                KDF = AuthDataHelper.GetSizedByteArray(pubArea, ref offset, 2);
            }

            // TPMU_PUBLIC_ID
            Unique = AuthDataHelper.GetSizedByteArray(pubArea, ref offset);
            if (pubArea.Length != offset)
                throw new VerificationException("Leftover bytes decoding pubArea");
        }
        public byte[] Raw { get; private set; }
        public byte[] Type { get; private set; }
        public byte[] Alg { get; private set; }
        public byte[] Attributes { get; private set; }
        public byte[] Policy { get; private set; }
        public byte[] Symmetric { get; private set; }
        public byte[] Scheme { get; private set; }
        public byte[] KeyBits { get; private set; }
        public uint Exponent { get; private set; }
        public byte[] CurveID { get; private set; }
        public byte[] KDF { get; private set; }
        public byte[] Unique { get; private set; }
        public TpmEccCurve EccCurve => (TpmEccCurve)Enum.Parse(typeof(TpmEccCurve), BitConverter.ToUInt16(CurveID.Reverse().ToArray(), 0).ToString());
        public ECPoint ECPoint
        {
            get
            {
                var point = new ECPoint();
                var uniqueOffset = 0;
                var size = AuthDataHelper.GetSizedByteArray(Unique, ref uniqueOffset, 2);
                point.X = AuthDataHelper.GetSizedByteArray(Unique, ref uniqueOffset, BitConverter.ToUInt16(size.Reverse().ToArray(), 0));
                size = AuthDataHelper.GetSizedByteArray(Unique, ref uniqueOffset, 2);
                point.Y = AuthDataHelper.GetSizedByteArray(Unique, ref uniqueOffset, BitConverter.ToUInt16(size.Reverse().ToArray(), 0));
                return point;
            }
        }
    }
}
