﻿//******************************************************************************************************************************************************************************************//
// Copyright (c) 2021 abergs (https://github.com/abergs/fido2-net-lib)                                                                                                                      //                        
//                                                                                                                                                                                          //
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),                                       //
// to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,   //
// and to permit persons to whom the Software is furnished to do so, subject to the following conditions:                                                                                   //
//                                                                                                                                                                                          //
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.                                                           //
//                                                                                                                                                                                          //
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,                                      //
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,                            //
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               //
//                                                                                                                                                                                          //
//******************************************************************************************************************************************************************************************//
using System;
using System.IO;
using System.Runtime.InteropServices;

namespace Neos.IdentityServer.MultiFactor.WebAuthN.Objects
{
    public class AttestedCredentialData
    {
        /// <summary>
        /// Minimum length of the attested credential data structure.  AAGUID + credentialID length + credential ID + credential public key.
        /// <see cref="https://www.w3.org/TR/webauthn/#attested-credential-data"/>
        /// </summary>
        private readonly int _minLength = Marshal.SizeOf(typeof(Guid)) + sizeof(ushort) + sizeof(byte) + sizeof(byte);

        /// <summary>
        /// The AAGUID of the authenticator. Can be used to identify the make and model of the authenticator.
        /// <see cref="https://www.w3.org/TR/webauthn/#aaguid"/>
        /// </summary>
        public Guid AaGuid { get; private set; }

        /// <summary>
        /// A probabilistically-unique byte sequence identifying a public key credential source and its authentication assertions.
        /// <see cref="https://www.w3.org/TR/webauthn/#credential-id"/>
        /// </summary>
        public byte[] CredentialID { get; private set; }

        /// <summary>
        /// The credential public key encoded in COSE_Key format, as defined in 
        /// Section 7 of RFC8152, using the CTAP2 canonical CBOR encoding form.
        /// <see cref="https://www.w3.org/TR/webauthn/#credential-public-key"/>
        /// </summary>
        public CredentialPublicKey CredentialPublicKey { get; private set; }

        internal static void SwapBytes(byte[] bytes, int index1, int index2)
        {
            var temp = bytes[index1];
            bytes[index1] = bytes[index2];
            bytes[index2] = temp;
        }

        /// <summary>
        /// AAGUID is sent as big endian byte array, this converter is for little endian systems.
        /// </summary>
        public static Guid FromBigEndian(byte[] Aaguid)
        {
            SwapBytes(Aaguid, 0, 3);
            SwapBytes(Aaguid, 1, 2);
            SwapBytes(Aaguid, 4, 5);
            SwapBytes(Aaguid, 6, 7);

            return new Guid(Aaguid);
        }

        /// <summary>
        /// AAGUID is sent as big endian byte array, this converter is for little endian systems.
        /// </summary>
        public static byte[] AaGuidToBigEndian(Guid AaGuid)
        {
            var aaguid = AaGuid.ToByteArray();

            SwapBytes(aaguid, 0, 3);
            SwapBytes(aaguid, 1, 2);
            SwapBytes(aaguid, 4, 5);
            SwapBytes(aaguid, 6, 7);

            return aaguid;
        }

        /// <summary>
        /// Instantiates an AttestedCredentialData object from an aaguid, credentialID, and CredentialPublicKey
        /// </summary>
        /// <param name="aaguid"></param>
        /// <param name="credentialID"></param>
        /// <param name="cpk"></param>
        public AttestedCredentialData(Guid aaguid, byte[] credentialID, CredentialPublicKey cpk)
        {
            AaGuid = aaguid;
            CredentialID = credentialID;
            CredentialPublicKey = cpk;
        }

        public AttestedCredentialData(byte[] bytes) : this(new BinaryReader(new MemoryStream(bytes, false)))
        {
        }

        /// <summary>
        /// Decodes attested credential data.
        /// </summary>
        public AttestedCredentialData(BinaryReader reader)
        {
            if (reader.BaseStream.Length < _minLength)
                throw new VerificationException("Not enough bytes to be a valid AttestedCredentialData");
            // First 16 bytes is AAGUID
            var aaguidBytes = reader.ReadBytes(Marshal.SizeOf(typeof(Guid)));

            if (BitConverter.IsLittleEndian)
            {
                // GUID from authenticator is big endian. If we are on a little endian system, convert.
                AaGuid = FromBigEndian(aaguidBytes);
            }
            else
            {
                AaGuid = new Guid(aaguidBytes);
            }

            // Byte length of Credential ID, 16-bit unsigned big-endian integer. 
            var credentialIDLenBytes = reader.ReadBytes(sizeof(ushort));

            if (BitConverter.IsLittleEndian)
            {
                // Credential ID length from authenticator is big endian.  If we are on little endian system, convert.
                Array.Reverse(credentialIDLenBytes);
            }

            // Convert the read bytes to uint16 so we know how many bytes to read for the credential ID
            var credentialIDLen = BitConverter.ToUInt16(credentialIDLenBytes, 0);

            // Read the credential ID bytes
            CredentialID = reader.ReadBytes(credentialIDLen);

            // "Determining attested credential data's length, which is variable, involves determining 
            // credentialPublicKey's beginning location given the preceding credentialId's length, and 
            // then determining the credentialPublicKey's length"

            // Read the CBOR object from the stream
            CredentialPublicKey = new CredentialPublicKey(reader.BaseStream);
        }

        public override string ToString()
        {
            return $"AAGUID: {AaGuid}, CredentialID: {CredentialID.ToString().Replace("-", "")}, CredentialPublicKey: {CredentialPublicKey}";
        }

        public byte[] ToByteArray()
        {
            using (var ms = new MemoryStream())
            {
                using (var writer = new BinaryWriter(ms))
                {
                    // Write the aaguid bytes out, reverse if we're on a little endian system
                    if (BitConverter.IsLittleEndian)
                    {
                        writer.Write(AaGuidToBigEndian(AaGuid));
                    }
                    else
                    {
                        writer.Write(AaGuid.ToByteArray());
                    }

                    // Write the length of credential ID, as big endian bytes of a 16-bit unsigned integer
                    var credentialIDLen = (ushort)CredentialID.Length;
                    var credentialIDLenBytes = BitConverter.GetBytes(credentialIDLen);
                    if (BitConverter.IsLittleEndian)
                    {
                        Array.Reverse(credentialIDLenBytes);
                    }

                    writer.Write(credentialIDLenBytes);

                    // Write CredentialID bytes
                    writer.Write(CredentialID);

                    // Write credential public key bytes
                    writer.Write(CredentialPublicKey.GetBytes());
                }
                return ms.ToArray();
            }
        }
    }
}
